#ifndef __3D_MATH_HPP
#define __3D_MATH_HPP
/*/ 3DMATH.HPP
 Basic structures and routines for 3D code, but with a more
 OOP twist to it....
 Smashed together by Kent "MenThal" Dahl
*/
#define vector2D_datatype int
#include "vector.hpp" //"kdiv\include\vector.hpp"

#define APPROXIMATION 0.0001  //allowed difference between FP values which still are treated as EQUAL


//OLD STUFF
//Here follows the stuff for our COS and SIN

boolean inline ApproxEqual(FP_type f1, FP_type f2)
{
	if( fabs(f1-f2)<APPROXIMATION ) return TRUE;
	return FALSE;
}



typedef WORD angle_t;
#define AngleRange 4096
FP_type sintab[AngleRange];

void Math3D_SetUp_SineTable()
{ int i;
 for(i=0;i<AngleRange;i++) sintab[i] = sin(M_PI * i * 2 / AngleRange);
}

inline FP_type Sin(angle_t a)
{	return sintab[int(a)&(AngleRange-1)];	}
inline FP_type Cos(angle_t a)
{	return sintab[(int(a)+(AngleRange>>2))&(AngleRange-1)];	}

//#define Cos(x) sintab[(int(x)+(AngleRange>>2))&(AngleRange-1)]


//The IMPORTANT datatypes: CLASSES.....children of vectors...
class point2D_t : public vector2D
{
	public:
		point2D_t();
		point2D_t(vector2D);
		point2D_t(vector2D_datatype, vector2D_datatype);
		friend boolean ApproxEqual(point2D_t&, point2D_t&);
		vector2D_dt V(int);

};

class point_t : public vector3D
{
	public:
		point_t();
		point_t(vector3D);
		point_t(FP_type, FP_type, FP_type);
		point2D_t ProjectTo2D();
		friend boolean ApproxEqual(point_t&, point_t&);
		//FP_type V(int);

};


class rotation3D_t    //Poorly tested...
{
	protected:
		angle_t a[3];
	public:
		rotation3D_t();
		rotation3D_t(angle_t, angle_t, angle_t);
		FP_type GetSinOf(unsigned int);
		FP_type GetCosOf(unsigned int);
		void Rotate(WORD, int);
};

void rotation3D_t::Rotate(WORD ax, int deg){
	a[ax%3]+=deg*AngleRange/360;
}

class pMatrix_t     //Matrix for projecting points/vectors onto new axes/axises..
{                   //Poorly tested...
	protected:
		point_t p[3];
	public:
		pMatrix_t();
		pMatrix_t(point_t, point_t, point_t);
		void SetToRotatedAxes(rotation3D_t&);
		point_t UsedOn(vector3D);
};

class Position6D_t : public point_t, public rotation3D_t {
	public:
		Position6D_t();
		Position6D_t(point_t);
		Position6D_t(point_t, rotation3D_t);
};



// The definition of the above classes' functions....
//point2D_t
point2D_t::point2D_t()  : vector2D()
	{}
point2D_t::point2D_t(vector2D_datatype v0, vector2D_datatype v1) : vector2D(v0, v1)
	{}
point2D_t::point2D_t(vector2D V) : vector2D(V)    //Ctor for reuse of vector2D functions.
	{}
inline vector2D_dt point2D_t::V(int i){
	return v[i];	}

/*boolean ApproxEqual(point2D_t &V1, point2D_t &V2)
{
	if( ApproxEqual(V1.v[0], V2.v[0])
		&& ApproxEqual(V1.v[1], V2.v[1]) )
		 { return TRUE;}
		return FALSE;
} */

//...point2D_t


//point_t
point_t::point_t() : vector3D()         //Ctor
	{}
point_t::point_t(vector3D V) : vector3D(V)    //Ctor for reuse of vector3D functions.
	{}
point_t::point_t(FP_type v0, FP_type v1, FP_type v2)  //Ctor for param passing to 
	: vector3D(v0, v1, v2)	{}

point2D_t point_t::ProjectTo2D() //3D -> 2D projection...not finished
{   FP_type tmp;
	tmp = 256.0/v[2];
	return point2D_t(v[0]*tmp, v[1]*tmp);
}

boolean ApproxEqual(point_t &V1, point_t &V2)
{
	if(ApproxEqual(V1.v[0], V2.v[0])
		&& ApproxEqual(V1.v[1], V2.v[1])
		&& ApproxEqual(V1.v[2], V2.v[2])
		)   { return TRUE;}
		return FALSE;
}

//...point_t

//rotation3D_t
rotation3D_t::rotation3D_t()   //Ctor
	{ a[0]=a[1]=a[2]=0;	}
rotation3D_t::rotation3D_t(angle_t a0, angle_t a1, angle_t a2) //Ctor w/ param
	{	a[0]=a0; a[1]=a1; a[2]=a2;	}

FP_type rotation3D_t::GetSinOf(unsigned int i)
	{	if(i<3)	return( Sin(a[i]));
		else return 0;	}
FP_type rotation3D_t::GetCosOf(unsigned int i)
	{	if(i<3) return Cos(a[i]);
		else return 0;	}

//....rotation3D_t

//pMatrix_t
pMatrix_t::pMatrix_t()	{	}		//Ctor
pMatrix_t::pMatrix_t(point_t p0, point_t p1, point_t p2)	//Ctor w/ param.
	{	p[0]=p0; p[1]=p1; p[2]=p2;	}

void pMatrix_t::SetToRotatedAxes(rotation3D_t &R)
{
	FP_type sX,sY,sZ,cX,cY,cZ;
		//pMatrix_t Mat;
	sX = R.GetSinOf(0);		cX = R.GetCosOf(0);
	sY = R.GetSinOf(1);		cY = R.GetCosOf(1);
	sZ = R.GetSinOf(2);		cZ = R.GetCosOf(2);

	//Make the matrix contain new axis all at once...
	p[0] = point_t( (cY*cZ),			(sZ),			(-sY*cZ) );
	p[1] = point_t( (sY*sX-cY*sZ*cX),	(cZ*cX),	(sY*sZ*cX+cY*sX) );
	p[2] = point_t( (sY*cX+cY*sZ*sX),	(-cZ*sX),	(-(sY*sZ*sX-cY*cX)));
}
point_t pMatrix_t::UsedOn(vector3D P){
	return point_t( P*p[0], P*p[1], P*p[2]);
}
//.....pMatrix_t

//Position6D_t
Position6D_t::Position6D_t() : point_t(), rotation3D_t(){
}
Position6D_t::Position6D_t(point_t p) : point_t(p){
}
Position6D_t::Position6D_t(point_t p, rotation3D_t r)
		 : point_t(p), rotation3D_t(r){
}
//Position6D_t(Position6D_t p) : point_t(p), rotation3D_t(p){}

point2D_t ScreenCenterPoint;

//.....Position6D_t


/* //Ol' crap stuff...
typedef struct {  //3D vector type... 3D point
	FP_type v[3];
} point_t;

typedef struct {  //2D point
	int x, y;
} point2D_t;

#define DotProduct(x,y) (x.v[0]*y.v[0]+x.v[1]*y.v[1]+x.v[2]*y.v[2])
void CrossProduct(point_t *U, point_t *V, point_t *UxV)
{
 UxV->v[0] = (U->v[1] * V->v[2]) - (U->v[2] * V->v[1]);
 UxV->v[1] = (U->v[2] * V->v[0]) - (U->v[0] * V->v[2]);
 UxV->v[2] = (U->v[0] * V->v[1]) - (U->v[1] * V->v[0]);
}

int GetNormal(point_t *U, point_t *V, point_t *Normal)
//Calc the normal of the plane denoted by vectors U & V
{
 FP_type temp;
 CrossProduct(U,V,Normal);
 temp = sqrt(Normal->v[0]*Normal->v[0]
			  + Normal->v[1]*Normal->v[1] + Normal->v[2]*Normal->v[2]);
 if(fabs(temp)<APPROXIMATION) return (-1);
 temp=1/temp;
 Normal->v[0]*=temp;
 Normal->v[1]*=temp;
 Normal->v[2]*=temp;
 return(0);
}
*/

#endif