// Flat Shading 3D Engine source - FS3DE.CPP - by Kent Dahl
/* O.O.P. approach
Start:	28 / 02 - 1997
Update: 09 / 03 - 1997
- Supposed to contain the engine, without too much hardware specific stuff
*/
#ifndef __FS3DE_CPP
#define __FS3DE_CPP

#include "fs3de.hpp"


Face_t::Face_t(ListIndex_t v0, ListIndex_t v1, ListIndex_t v2,
		ListIndex_t n, ColorIndex_t col) : normal(n), color(col)
{		fv[0]=v0;  fv[1]=v1;   fv[2]=v2;	}
/////////////////////////


Face2D_t::Face2D_t() : color(0){};
Face2D_t::Face2D_t(ListIndex_t fp0, ListIndex_t fp1, ListIndex_t fp2,
									 ColorIndex_t c) : color(c)
{	fp[0]=fp0; fp[1]=fp1; fp[2]=fp2; }

/*boolean inline Face2D_t::operator >(Face2D_t F2){
	return ((depth>F2.depth) ? TRUE : FALSE );
} */


/////////////////////////
VertexList::VertexList(ListIndex_t n) : SimpleList<point_t>(n){}
ScreenCoordsList::ScreenCoordsList(ListIndex_t n) : SimpleList<point2D_t>(n){};
FaceList::FaceList(ListIndex_t n) : SimpleList<Face_t>(n){};
NormalList::NormalList(ListIndex_t n) : SimpleList<point_t>(n){};
Face2DList::Face2DList(ListIndex_t n) : SimpleList<Face2D_t>(n){};

SimpleObjData::SimpleObjData(VertexList *v, NormalList *n, FaceList *f )
				 : VL(v), NL(n), FL(f) {}


inline DepthSortIndex::DepthSortIndex() : I(0), Z(0) {};
inline DepthSortIndex::DepthSortIndex(ListIndex_t i, FP_type z)
			 : I(i), Z(z) {};
inline boolean DepthSortIndex::operator >(DepthSortIndex D2)
			{ return (Z>D2.Z ? TRUE : FALSE ); }

DepthSortIndexList::DepthSortIndexList(WORD n) : SortableList<DepthSortIndex>(n) { };

void DepthSortIndexList::AddSorted(DepthSortIndex DSI){ //Untested
	int i=0;
	do{
	}while(i<Num && V[i]>DSI);
	for(int j=Num;j>i;j++) V[j]=V[j-1];
	V[i]=DSI;
	Num++;
}
////
WorldSpaceData_t::WorldSpaceData_t(){}
WorldSpaceData_t::WorldSpaceData_t(EntityList *e, WorldSpace *w) : EL(e), WS(w){}


/////////////////////////

ScreenSpace_t::ScreenSpace_t(ListIndex_t nsc, ListIndex_t nf2d){
	SCL = new ScreenCoordsList(nsc);
	F2DL = new Face2DList(nf2d);
	DSIL = new DepthSortIndexList(nf2d);
}
ScreenSpace_t::ScreenSpace_t(){
	SCL = new ScreenCoordsList(65536/sizeof(point2D_t));
	F2DL = new Face2DList(65536/sizeof(Face2D_t));
	DSIL = new DepthSortIndexList(65536/sizeof(Face2D_t));
}
void ScreenSpace_t::Clear(){
	SCL->EmptyList();
	F2DL->EmptyList();
	DSIL->EmptyList();
}



FlatShading3DEngine* _Current_FS3DE;  //Currently active FS3DE; global var. passing
FlatShading3DEngine::FlatShading3DEngine()
{
	ScreenSpace = new ScreenSpace_t;
	//TmpVL = new VertexList(??);
	//TmpNL = new NormalList(??);
	SetCurrent();
}

void FlatShading3DEngine::SetCurrent(){
	_Current_FS3DE = this;
}
void FlatShading3DEngine::Draw(){
	ScreenSpace->Draw();
}
void FlatShading3DEngine::ClearOutput(){
	ScreenSpace->Clear();
}


//FlatShading3DEngine::

// Process vertexes - Should NOT contain ANY hardware, nor
// project, specific code
void SimpleObjData::AddToScreenSpaceAt(Position6D_t ObjPos){
	VertexList tmpVL(VL->NumOfElems()); //=_Current_FS3DE->TmpVL;//;
	NormalList tmpNL(NL->NumOfElems()); //=_Current_FS3DE->TmpNL;//;
	pMatrix_t ObjRotMat, CameraRotMat;
	ScreenCoordsList* ScrCL = _Current_FS3DE->ScreenSpace->SCL;  //ptr;
	Face2DList* F2DL = _Current_FS3DE->ScreenSpace->F2DL;
	point_t P;					point2D_t p;    FP_type Shade, XShade, Depth;
	Position6D_t* pScrCam = _Current_FS3DE->ScreenSpace->Camera;
	ListIndex_t SCLStart = ScrCL->NumOfElems();

	ObjRotMat.SetToRotatedAxes(ObjPos);
	CameraRotMat.SetToRotatedAxes(*pScrCam);

	tmpVL.EmptyList();  //Do vertexList
	VL->ResetStart();
	do{
		P=ObjRotMat.UsedOn(VL->Current());
		P=CameraRotMat.UsedOn( P+ObjPos-(*pScrCam) );
		tmpVL.Add(P);
		//ScrCL->Add(( (P*vector3D(0,0,1)>0) ? P.ProjectTo2D() : point2D_t(-3000,-3000) ));
	}while(++(*VL));

	tmpNL.EmptyList();    //Rotate normalList
	NL->ResetStart();
	do{
		P=CameraRotMat.UsedOn( ObjRotMat.UsedOn( NL->Current() ));
		//P=P+ObjPos;
		tmpNL.Add(P);
	}while(++(*NL));

	tmpVL.ResetStart();   //Adds the points to list of screenspace coords
	do{
		P=tmpVL.Current();
		if(P*vector3D(0,0,1)>0){
			ScrCL->Add((P.ProjectTo2D()+ScreenCenterPoint));
		} else ScrCL->Add(point2D_t(-3000,-3000));

	}while(++tmpVL);   /* */

	FL->ResetStart();
	do{
		Shade = tmpVL[FL->Current().fv[0]]*tmpNL[(FL->Current().normal)];
		if(Shade>0) //Backface culling...
		{
			Shade=tmpNL[(FL->Current().normal)].Dot(vector3D(0,0,1))*0.4999+0.5;
			Depth = (tmpVL[FL->Current().fv[0]]+
							 tmpVL[FL->Current().fv[1]]+
							 tmpVL[FL->Current().fv[2]])*vector3D(0,0,1)/3;
			_Current_FS3DE->ScreenSpace->
						DSIL->Add( DepthSortIndex(F2DL->NumOfElems(), Depth) );
			if(Depth>50) Shade = Shade * 50 / Depth;

			F2DL->Add( Face2D_t( FL->Current().fv[0]+SCLStart,
												FL->Current().fv[1]+SCLStart,
												FL->Current().fv[2]+SCLStart,
												(BYTE) ((FL->Current().color)*NUM_SHADES+NUM_SHADES*Shade )
											 ));
		}
	}while(++(*FL));
	_Current_FS3DE->ScreenSpace->DSIL->SortQ();
}



#endif //End-of-file