#ifndef __VECTOR_HPP
#define __VECTOR_HPP
// VECTOR.HPP - 3D Vector CLASS and routines by Kent Dahl, "MenThal"
/* Start: 01.Feb.1997 	- Update 09.02.97

 A OOP approach to the datastructures used in the 3D-project (Remember
 POINT_T datastructure in Abrash's example???)
 - I'm hoping to create a more solid base for my 3D-projects, thus
not making a mess of my next one. NOTE: Use inheritance instead of extending the
current vector-classes. A point2D/3D_t class could INHERIT from the vector2D/3D class
and extend its more application specific functions from there(as f.ex. Project3DTo2D..)

 - Maybe I ought to change the Dot and Cross for this class,
as they may be confusing as to where the results will go... Should the functions
that alter the vector return a vector or simply alter the contents directly???

CLASS: vector2D = 2D vector type made up of 2 FP values
 overloaded operators:
  + - * << >> ==

CLASS: vector3D	= 3D vector type made up of 3 FP values
 overloaded operators:
  + - * (add, subtract, dot(-vector?vector-)/multiply(-vector?FP_type-)
		 all defined (INLINE?)
	<< >> ==

- Functions should perhaps be virtual for inheritance purposes, although that may
slow execution speed down...
- I think the inheritance between vector2D and vector3D was a bad idea, and removed it.
O.O.P. inheritance should include more than just the basic data structures...

NOTE: Do not extend these classes for specific purposes (as 3D-projection, rotation etc.)
but create a new class, a Child of vector?D, and add the capabilities there...

*/
#ifndef __cplusplus
#error Must use C++ for the classes...:-)
#endif


#include <math.h>
#include <iostream.h>
#include "common.h"

//vector2D class
/* Perhaps the functions in this class, that can be redefined , should have
been virtual... I don't know... But I'm avoiding LATE LINKAGE as much as I can until
I fully comprehend it...
*/

#define vector2D_dt int
class vector2D
{
	protected:
		vector2D_dt v[2];
	public:
		vector2D();
		vector2D(vector2D_dt, vector2D_dt);
		vector2D_dt Length();		//Slow length formula
		vector2D_dt Dot(vector2D);
		void Unitise();			//Divide each by length(slow one)

		vector2D operator + (vector2D);		//Addition
		vector2D operator - (vector2D);		//Subtraction
		vector2D_dt operator * (vector2D);		//Dotproduct..
		vector2D operator * (vector2D_dt);  //Multiply each/lengthen by..
		boolean operator == (vector2D);			//Logical equal

		friend ostream& operator << (ostream&, vector2D&);		//Print out: X Y
		friend istream& operator >> (istream&, vector2D&);		//Get in:   X Y
};

	//Functions..
 vector2D::vector2D(){ v[0]=v[1]=0; }   //Constructors
 vector2D::vector2D( vector2D_dt v0, vector2D_dt v1)	{ v[0]=v0; v[1]=v1; }

 vector2D_dt vector2D::Length()	//Return length of vector
	{	return sqrt( v[0]*v[0]+v[1]*v[1] );}
 vector2D_dt vector2D::Dot(vector2D B) //Use * instead for speed...
	{	return (v[0]*B.v[0]+v[1]*B.v[1]);}
 void vector2D::Unitise()     //Make vector a unit long...
	{ 	vector2D_dt len = Length();
		if(len!=0) { len=1/len; v[0]*=len; v[1]*=len; }
	}

 inline vector2D vector2D::operator + (vector2D B)		//Addition
	{return vector2D(v[0]+B.v[0], v[1]+B.v[1]);}
 inline vector2D vector2D::operator - (vector2D B)		//Subtraction
	{return vector2D(v[0]-B.v[0], v[1]-B.v[1]);}
 inline vector2D_dt vector2D::operator * (vector2D B)		//Dotproduct..
	{return ( v[0]*B.v[0]+v[1]*B.v[1] );}
 inline vector2D vector2D::operator * (vector2D_dt B)  //Multiply each/lengthen by..
	{return vector2D(B*v[0], B*v[1]);}
 inline boolean vector2D::operator == (vector2D B)			//Logical equal
	{if( (v[0]==B.v[0]) && (v[1]==B.v[1]) )return TRUE; else return FALSE; }

 ostream& operator << (ostream& Os, vector2D& V)
	{	Os << " "<<V.v[0]<<" "<<V.v[1]<<" "; return Os; }
 istream& operator >> (istream& Is, vector2D& V)
	{	Is >> V.v[0] >> V.v[1]; return Is;}

//End of vector2D class




/*vector3D class;  based on the point_t-type, NOT on the vector2D class, as this
mainly just inherits the data type... The idea of inheritance and O.O.P. go beyond that
into the realm of functions...
*/
class vector3D //: public vector2D
{
	protected:
		FP_type v[3];
	public:
		vector3D();
		vector3D(FP_type, FP_type, FP_type);
		FP_type Length();		//Slow length formula
		FP_type Dot(vector3D);  //Use * instead...
		vector3D Cross(vector3D);
		void Unitise();			//Divide each by length(slow one)

		vector3D operator + (vector3D); //Addition
		vector3D operator - (vector3D); //Subtraction
		FP_type  operator * (vector3D);  //Dotproduct..
		vector3D operator * (FP_type);  //Multiply each/lengthen by..
		boolean operator == (vector3D);  //Logical equal

		friend ostream& operator << (ostream&, vector3D&); //Print out: X Y Z
		friend istream& operator >> (istream&, vector3D&); //Get in:   X Y Z
};

	//Functions..
vector3D::vector3D() //: vector2D(){ v[2]=0; }   //Constructors
{ v[0]=v[1]=v[2]=0; }
vector3D::vector3D( FP_type v0, FP_type v1, FP_type v2)// : vector2D(v0, v1)
	{v[0]=v0; v[1]=v1; v[2]=v2; }
FP_type vector3D::Length()	//Return length of vector
	{ return sqrt( v[0]*v[0]+v[1]*v[1]+v[2]*v[2] );}
FP_type vector3D::Dot(vector3D B) //Use * instead for speed...
	{	return ( v[0]*B.v[0]+v[1]*B.v[1]+v[2]*B.v[2]);}
vector3D vector3D::Cross(vector3D B)
	{	return vector3D( v[1]*B.v[2]-v[2]*B.v[1], v[2]*B.v[0]-v[0]*B.v[2],
		v[0]*B.v[1]-v[1]*B.v[0] );		}
void vector3D::Unitise()     //Make vector a unit long...
	{ 	FP_type len = Length();
		if(len!=0) { len=1/len; v[0]*=len; v[1]*=len; v[2]*=len; }	}

inline vector3D vector3D::operator + (vector3D B) //Addition
	{return vector3D(v[0]+B.v[0], v[1]+B.v[1], v[2]+B.v[2]);}
inline vector3D vector3D::operator - (vector3D B) //Subtraction
	{return vector3D(v[0]-B.v[0], v[1]-B.v[1], v[2]-B.v[2]);}
inline FP_type vector3D::operator * (vector3D B)  //Dotproduct..
	{return (v[0]*B.v[0] + v[1]*B.v[1] + v[2]*B.v[2]);}
inline vector3D vector3D::operator * (FP_type B)  //Multiply each/lengthen by..
	{return vector3D(B*v[0], B*v[1], B*v[2]);}
inline boolean vector3D::operator == (vector3D B)  //Logical equal
	{	if( (v[0]==B.v[0]) && (v[1]==B.v[1]) && (v[2]==B.v[2]) )
		return TRUE; else return FALSE;	}

ostream& operator << (ostream& Os, vector3D& V)
	{ Os<<V.v[0]<<" "<<V.v[1]<<" "<<V.v[2]<<" "; return Os; } //Inheritance?? OLD: //{ Os << " "<<V.v[0]<<" "<<V.v[1]<<" "<<V.v[2]<<" "; return Os; }
istream& operator >> (istream& Is, vector3D& V)
	{ Is >> V.v[0] >> V.v[1] >> V.v[2]; return Is;}

//End of vector3D class


vector3D GetNormal(vector3D V1, vector3D V2)
{
	vector3D Norm;
	Norm=V1.Cross(V2);
	Norm.Unitise();  
	return Norm; 
}


//---------- Cut here for VECTOR.HPP ----------

#endif